package jp.co.riiiver.sdk.sample

import android.content.Context
import android.content.Intent
import android.graphics.Bitmap
import android.net.Uri
import android.view.View
import android.webkit.CookieManager
import android.webkit.WebView
import androidx.browser.customtabs.CustomTabsIntent
import com.riiiver.auth.*
import com.riiiver.sdk.ersdk.*
import jp.co.riiiver.sdk.sample.RiiiverSDKConfig.getDeviceIdArray
import jp.co.riiiver.sdk.sample.RiiiverSDKConfig.getDeviceIdList
import jp.co.riiiver.sdk.sample.RiiiverSDKConfig.pieceCorePackageName
import timber.log.Timber
import java.util.*


object RiiiverSDKManager : ERSDKDelegate {

    private var erSdk: ERSDK? = null

    private var accessToken: String? = null
    private var idToken: String? = null

    /** 認証済みかどうか (true:認証済み、false:未認証)
     *  Is this user authenticated?  true:  Yes,  false: No
     */
    val authenticated: Boolean
        get() {
            return ERAuth.getInstance().isLoggedIn
        }

    /**
     *  RiiiverSDKを初期化する
     *  @param context Activityのコンテキスト
     *  @param complete 認証済みかどうか (true:認証済み、false:未認証)
     */
    fun initialize(context: Context, complete: (isLoggedIn: Boolean) -> Unit) {
        Timber.i("initialize")
        if (erSdk != null) {
            // initialize 完了済み
            Timber.d("Initialize has already been completed")
            complete(authenticated)
            return
        }

        // ERAuthの初期化
        ERAuth.getInstance().initialize(context)

        erSdk = ERSDK.Builder.with(context)
            .includeBlockExecutorPackageName(pieceCorePackageName).build()

        // iiideaの実行状態を監視
        erSdk!!.setDelegate(this)

        // Preference関連の初期化
        IiideaSettingFragment.initialize(context)

        complete(authenticated)
    }

    /**
     *  RiiiverSDKからWeb認証画面表示する
     *  set a web view to be  invoked by Riiiver SDK for the web page authentication.
     */
    fun startWebLogin(delegate: IERAuthWebLoginCallback) {
        if (erSdk == null) {
            // initialize 未完了
            Timber.e("not initialized")
            delegate.onError(null)
            return
        }

        // ログイン画面を起動
        ERAuth.getInstance().startWebLogin(getDeviceIdList(), delegate)
    }

    /**
     *  RiiiverSDKにトークン情報を設定する
     */
    fun challengeAuthByWebLogin(intent: Intent, delegate: IERAuthChallengeCallback) {
        ERAuth.getInstance().challengeAuthByWebLogin(intent, delegate)
    }

    /**
     *  RiiiverSDKから認証トークンを取得して保持する
     *  Get the auth token and cache it.
     *  @isSuccess true > Success to get the authentication info.  false > Failed to get the auth info.  認証情報の取得が成功した場合はtrue、失敗した場合はfalse
     *  @error the error information from Riiiver SDK.  -   RiiiverSDKから返却されたerror情報
     */
    fun getAWSToken(completion: (isSuccess: Boolean, error: Exception?) -> Unit) {
        val erSdk = erSdk
        if (erSdk == null) {
            // initialize 未完了
            Timber.e("not initialized")
            completion(false, null)
            return
        }

        if (!ERAuth.getInstance().isLoggedIn) {
            //未認証の場合、認証トークンを取得できないので終了 - Nothing to do... just return.
            Timber.e("getAWSToken failed. Unauthenticated")
            completion(false, null)
            return
        }

        // RiiiverSDKから認証トークンを取得  -  Get the Auth token.
        ERAuth.getInstance().getToken(object : IERAuthGetTokenCallback {
            override fun onSuccess(token: ERAuthToken?) {
                Timber.i("getToken onSuccess.")

                if (token == null) {
                    Timber.e("getAWSToken failed. No token.")
                    completion(false, null)
                    return
                }

                if (token.accessToken == null) {
                    Timber.e("getAWSToken failed. No access-token.")
                    completion(false, null)
                    return
                }

                if (token.idToken == null) {
                    Timber.e("getAWSToken failed. No id-token.")
                    completion(false, null)
                    return
                }

                accessToken = token.accessToken
                idToken = token.idToken
                completion(true, null)
            }

            override fun onError(error: Exception?) {
                Timber.e("getToken onError.")
                completion(false, error)
            }
        })
    }

    /**
     * RiiiverSDKのセッション終了してログアウトする
     * End the session of the Riiiver authentication.  (Logout from Riiiver)
     */
    fun logout(completion: (isSuccess: Boolean, error: Exception?) -> Unit) {
        val erSdk = erSdk
        if (erSdk == null) {
            // initialize 未完了
            Timber.e("not initialized")
            completion(false, null)
            return
        }
        // セッションが切れないように毎回トークンを更新  -  Update the token to keep the auth session.
        getAWSToken { isSuccess: Boolean, error: Exception? ->
            if (!isSuccess) {
                Timber.e(
                    "updateAWSToken failed. reason: %s",
                    if (error != null) error.message else ""
                )
                completion(false, null)
                return@getAWSToken
            }
            ERAuth.getInstance().logout(object : IERAuthCallback {
                override fun onSuccess() {
                    Timber.i("logout succeeded.")
                    completion(true, null)
                }

                override fun onCancel() {
                    Timber.e("onCancel in")
                }

                override fun onError(error: java.lang.Exception?) {
                    Timber.e("logout onError. %s", error.toString())
                    completion(false, error)
                }
            })
        }
    }

    /**
     *  ログインユーザーのプロフィール情報をwebViewに表示する
     *  Display Riiiver Profile page via Riiiver SDK.
     */
    fun openProfile(webView: WebView, delegate: IERAuthProfileWebViewDelegate) {
        // SDKの問題で一時的にcookieを削除する必要がある(SDKが対処された後は削除予定)
        val cookieManager: CookieManager = CookieManager.getInstance()
        cookieManager.removeAllCookies(null)

        webView.setLayerType(View.LAYER_TYPE_HARDWARE, null)
        val deviceIds = getDeviceIdList()
        ERAuth.getInstance().openProfile(webView,  deviceIds, delegate)
    }

    /**
     *  RiiiverSDKからiiideaStore画面表示する
     *  Display Riiiver Store page via Riiiver SDK.
     */
    fun openStore(webView: WebView, delegate: IERAuthStoreWebViewDelegate) {
        // SDKの問題で一時的にcookieを削除する必要がある(SDKが対処された後は削除予定)
        val cookieManager: CookieManager = CookieManager.getInstance()
        cookieManager.removeAllCookies(null)

        webView.setLayerType(View.LAYER_TYPE_HARDWARE, null)
        val deviceIds = getDeviceIdList()
        ERAuth.getInstance().openStore(webView, deviceIds, delegate)
    }

    /**
     * RiiiverSDKから My iiidea 一覧を取得する
     * Get the my iiidea list from RiiiverSDK.
     *  @iiideaList  My iiidea 一覧. 取得に失敗した場合はnull
     *  @error the error information from Riiiver SDK.  -   RiiiverSDKから返却されたerror情報
     */
    fun getMyApplets(completion: (iiideaList: MutableList<ERApplet>?, error: Exception?) -> Unit) {
        val erSdk = erSdk
        if (erSdk == null) {
            // initialize 未完了
            Timber.e("not initialized")
            completion(null, null)
            return
        }

        // セッションが切れないように毎回トークンを更新  -  Update the token to keep the auth session.
        getAWSToken { isSuccess: Boolean, error: Exception? ->
            if (!isSuccess) {
                Timber.e(
                    "updateAWSToken failed. reason: %s",
                    if (error != null) error.message else ""
                )
                completion(null, error)
                return@getAWSToken
            }

            erSdk.getMyApplets(getDeviceIdArray()) { appletList: MutableList<ERApplet>?, erError: ERError? ->
                // Note: エラーがあっても取得できた My iiidea は返却されるため処理を継続することは可能 - Here, we can get the iiidea even if we get some errors.
                if (erError != null) {
                    Timber.e("getMyApplets failed. reason: %s", erError.message)
                }
                completion(appletList, null)
            }
        }
    }

    /**
     * 指定したiiideaをアクティベートする
     * Activate an terget iiidea
     *  @context  Activityのコンテキスト
     *  @iiidea  アクティベートするiiidea
     *  @isSuccess アクティベートが成功した場合はtrue、失敗した場合はfalse
     *  @error RiiiverSDKから返却されたerror情報
     */
    fun activate(
        context: Context,
        iiidea: ERApplet,
        completion: (isSuccess: Boolean, error: ERError?) -> Unit
    ) {
        val erSdk = erSdk
        if (erSdk == null) {
            // initialize 未完了
            Timber.e("not initialized")
            completion(false, null)
            return
        }

        // セッションが切れないように毎回トークンを更新  -  Update the token to keep the auth session.
        getAWSToken { isSuccess: Boolean, error: Exception? ->
            if (!isSuccess) {
                Timber.e(
                    "updateAWSToken failed. reason: %s",
                    if (error != null) error.message else ""
                )
                completion(false, null)
                return@getAWSToken
            }

            iiidea.activate(context) { _: Array<String>?, _: Array<ERServiceProxyData>?, erError: ERError? ->
                if (erError == null) {
                    Timber.d("activate succeeded.")
                    completion(true, null)
                } else {
                    completion(false, erError)
                }
            }
        }
    }

    /**
     * 指定したiiideaをディアクティベートする
     * Deactivate the target iiidea.
     *  @iiidea  ディアクティベートするiiidea  - target iiidea to deactivate.
     *  @isSuccess ディアクティベートが成功した場合はtrue、失敗した場合はfalse
     *  @error RiiiverSDKから返却されたerror情報
     */
    fun deactivate(iiidea: ERApplet, completion: (isSuccess: Boolean, error: Exception?) -> Unit) {
        val erSdk = erSdk
        if (erSdk == null) {
            // initialize 未完了
            Timber.e("not initialized")
            completion(false, null)
            return
        }

        // セッションが切れないように毎回トークンを更新  -  Update the token to keep the auth session.
        getAWSToken { isSuccess: Boolean, error: Exception? ->
            if (!isSuccess) {
                Timber.e(
                    "updateAWSToken failed. reason: %s",
                    if (error != null) error.message else ""
                )
                completion(false, error)
                return@getAWSToken
            }
            iiidea.deactivate()
            Timber.d("deactivate succeeded.")
            completion(true, null)
        }
    }

    /**
     * 指定したiiideaを削除する
     * Remove the target iiidea.
     *  @iiidea  削除するiiidea  - target iiidea to remove.
     *  @isSuccess 削除が成功した場合はtrue、失敗した場合はfalse
     *  @error RiiiverSDKから返却されたerror情報
     */
    fun removeMyIiidea(
        iiidea: ERApplet,
        completion: (isSuccess: Boolean, error: ERError?) -> Unit
    ) {
        val erSdk = erSdk
        if (erSdk == null) {
            // initialize 未完了
            Timber.e("not initialized")
            completion(false, null)
            return
        }

        // セッションが切れないように毎回トークンを更新  -  Update the token to keep the auth session.
        getAWSToken { isSuccess: Boolean, error: Exception? ->
            if (!isSuccess) {
                Timber.e(
                    "updateAWSToken failed. reason: %s",
                    if (error != null) error.message else ""
                )
                completion(false, null)
                return@getAWSToken
            }
            erSdk.removeMyApplet(iiidea) { erError: ERError? ->
                if (erError == null) {
                    completion(true, null)
                } else {
                    completion(false, erError)
                }
            }
        }
    }

    /**
     * SDKが生成したUUIDを取得する
     * Get the UUID.
     */
    fun getUuid(): UUID? {
        val erSdk = erSdk
        if (erSdk == null) {
            // initialize 未完了
            Timber.e("not initialized")
            return null
        }
        return erSdk.uuid
    }

    /**
     * SDKからPermissionの警告ダイアログを表示する
     */
    fun showPermissionDialog(
        context: Context,
        requestCode: Int,
        permissions: Array<out String>,
        grantResults: IntArray
    ): Boolean? {
        val erSdk = erSdk
        if (erSdk == null) {
            // initialize 未完了
            Timber.e("not initialized")
            return null
        }
        return erSdk.showPermissionsDialog(context, requestCode, permissions, grantResults)
    }

    /**
     * 認証のパーミッション許可画面を表示する。
     * Display a permission view to access an authentication.
     *
     * @blockId ID for target piece (block). - 対象ブロックのID。
     * @isSuccess 削除が成功した場合はtrue、失敗した場合はfalse
     */
    fun showAuthenticationSession(
        context: Context,
        blockId: String,
        completion: (isSuccess: Boolean, error: Exception?) -> Unit
    ) {
        // セッションが切れないように毎回トークンを更新  -  Update the token to keep the auth session.
        getAWSToken { isSuccess: Boolean, error: Exception? ->
            if (!isSuccess) {
                Timber.e(
                    "updateAWSToken failed. reason: %s",
                    if (error != null) error.message else ""
                )
                completion(false, error)
                return@getAWSToken
            }
            erSdk!!.getClientToken{ token: String?, erError: ERError? ->
                if (erError == null) {
                    Timber.d("getClientToken succeeded.")
                    val builder = CustomTabsIntent.Builder()
                    val customTabsIntent = builder.build()
                    val uri =
                        Uri.parse(RiiiverSDKConfig.riiiverAuthUri + "?blockId=" + blockId + "&token=" + token + "&uuid=" + getUuid())
                    customTabsIntent.launchUrl(context, uri)
                    completion(true, null)
                } else {
                    completion(false, Exception(erError.message))
                }
            }
        }
    }

    /**
     * 認証状態を解除する。
     *
     * @blockId ID for target piece (block). - 対象ブロックのID。
     * @isSuccess 削除が成功した場合はtrue、失敗した場合はfalse
     */
    fun stopAuthenticationSession(
        blockId: String,
        completion: (isSuccess: Boolean, error: ERError?) -> Unit
    ) {
        val erSdk = erSdk
        if (erSdk == null) {
            // initialize 未完了
            Timber.e("not initialized")
            completion(false, null)
            return
        }
        // セッションが切れないように毎回トークンを更新  -  Update the token to keep the auth session.
        getAWSToken { isSuccess: Boolean, error: Exception? ->
            if (!isSuccess) {
                Timber.e(
                    "updateAWSToken failed. reason: %s",
                    if (error != null) error.message else ""
                )
                completion(false, null)
                return@getAWSToken
            }
            erSdk.deleteClientAuth(blockId) { erError: ERError? ->
                if (erError == null) {
                    completion(true, null)
                } else {
                    completion(false, erError)
                }
            }

        }
    }

    /**
     * 指定したiiideaを強制的にダウンロードする
     */
    fun forcedDownloadApplet(completion: (isSuccess: Boolean, error: ERError?) -> Unit){
        val erSdk = erSdk
        if (erSdk == null) {
            // initialize 未完了
            Timber.e("not initialized")
            completion(false, null)
            return
        }

        // セッションが切れないように毎回トークンを更新  -  Update the token to keep the auth session.
        getAWSToken { isSuccess: Boolean, error: Exception? ->
            if (!isSuccess) {
                Timber.e(
                    "updateAWSToken failed. reason: %s",
                    if (error != null) error.message else ""
                )
                completion(false, null)
                return@getAWSToken
            }
            // 強制ダウンロード対象iiideaをダウンロード状態に設定する
            val appletIdList = RiiiverSDKConfig.getAppletIdList()
            forcedDownloadApplet(appletIdList) { isSuccess2: Boolean, erError: ERError? ->
                if (isSuccess2) {
                    completion(true, null)
                }
                else {
                    completion(false, erError)
                }
            }
        }
    }

    private fun forcedDownloadApplet(appletIdList: List<String>, completion: (isSuccess: Boolean, error: ERError?) -> Unit) {
        val erSdk = erSdk
        if (erSdk == null) {
            // initialize 未完了
            Timber.e("not initialized")
            completion(false, null)
            return
        }
        if (appletIdList.isEmpty()){
            // 対象のiiideaなし
            Timber.d("No forced download applet")
            completion(true, null)
            return
        }
        erSdk.forcedDownloadApplet(appletIdList) { erError: ERError? ->
            if (erError == null) {
                completion(true, null)
            } else {
                completion(false, erError)
            }
        }
    }

    /**
     * Riiiverからのお知らせを表示する
     *  @webView  お知らせページを表示するWebView
     */
    fun showNews(webView: WebView) {
        val erSdk = erSdk
        if (erSdk == null) {
            // initialize 未完了
            Timber.e("not initialized")
            return
        }
        erSdk.showNews(webView, getDeviceIdList())
    }

    /**
     * Riiiverからの未読のお知らせがあるか確認する
     * @isUnread 未読のお知らせがあるかどうか(trueの場合は未読のお知らせあり)
     * @error RiiiverSDKから返却されたerror情報
     */
    fun isUnreadNews(completion: (isUnread: Boolean, error: ERError?) -> Unit) {
        val erSdk = erSdk
        if (erSdk == null) {
            // initialize 未完了
            Timber.e("not initialized")
            completion(false, null)
            return
        }
        erSdk.isUnreadNews{ isUnread: Boolean, erError: ERError? ->
            completion(isUnread, erError)
        }
    }

    /**
     * 指定したURLの画像を取得する
     * Get the Image.
     * Deactivate the target iiidea.
     *  @imageUrl  取得する画像のURL
     *  @image 画像情報
     *  @s 画像情報
     *  @error RiiiverSDKから返却されたerror情報
     */
    fun getImage(
        imageUrl: String,
        completion: (image: Bitmap?, s: String?, error: ERError?) -> Unit
    ) {
        val erSdk = erSdk
        if (erSdk == null) {
            // initialize 未完了
            Timber.e("not initialized")
            completion(null, null, null)
            return
        }

        erSdk.getImage(imageUrl) { bitmap: Bitmap?, s: String?, erError: ERError? ->
            completion(bitmap, s, erError)
        }

    }

    /**
     * iiideaが実行開始した時に通知する
     * @param applet 起動したiiideaのインスタンス
     * @param error 起動に失敗した時のエラー
     */
    override fun startApplet(applet: ERApplet?, error: ERError?) {
        Timber.i("startApplet: error=${error}")
        if (applet == null) {
            Timber.e("startApplet: no iiidea")
            return
        }
        if (error != null) {
            Timber.e("startApplet: error %s", error.message)
            return
        }
        Timber.i("startApplet: appletId=${applet.id}, appletTitle=${applet.applet.title.get()}")
    }

    /**
     * Activate状態が更新された時に通知する
     *
     * - ERSDK.getMyApplet() の呼び出し後、更新のあった iiidea を返却するコールバック
     * - 更新前にActivate状態だったiiideaについては、iiidea更新後に再Activateを行う。
     * - 再Activateに失敗したiiideaについては、 deactivatedToError で iiidea の ID とエラー内容のMapの形でコールバックされる
     * @param updatedList 更新されたPieceのリスト
     * @param deactivatedToError RiiiverSDKから返却されたerror情報
     */
    override fun didUpdateAppletList(updatedList: MutableList<ERApplet>?, deactivatedToError: MutableMap<String, ERError>?) {
        Timber.i("didUpdateAppletList")
    }

    /**
     * iiideaが実行終了した時に通知する
     * @param applet 終了したiiideaのインスタンス
     * @param error iiideaが正しく終了しなかった時のエラー
     * @param reachBlock iiideaが終了した時の最終到達Block
     * @param output iiideaが正しく終了した際の、最後に到達したPieceが出力したデータ
     */
    override fun endApplet(
        applet: ERApplet?,
        error: ERError?,
        reachBlock: ReachBlock?,
        output: MutableMap<String, Any>?
    ) {
        Timber.i("endApplet: error=${error}, reachBlock=${reachBlock} output=${output}")
        if (applet == null) {
            Timber.e("endApplet: no iiidea")
            return
        }
        if (error != null) {
            Timber.e("endApplet: error %s", error.message)
            return
        }
        Timber.i("endApplet: appletId=${applet.id}, appletTitle=${applet.applet.title.get()}")
    }

}
